local addonName, addonTable = ...
-- Debugging to tekDebug
local debug = false
--[===[@debug@
debug = true
--@end-debug@]===]

local name, _, _, enabled = GetAddOnInfo("tekDebug")
local player = UnitName("player")
local enabled = GetAddOnEnableState(player, "tekDebug") == 2
if enabled and not IsAddOnLoaded("tekDebug") then
  local succ, err = LoadAddOn(name)
end
local debugf = debug and tekDebug and tekDebug:GetFrame(addonName)
local function DebugPrint(...) if debugf then debugf:AddMessage(string.join(", ", tostringall(...))) end end

local strfind, strformat, strgsub, strlower, strmatch, tostring = string.find, string.format, string.gsub, string.lower, strmatch, tostring
local _, date, ipairs, next, pairs, select, time, tinsert, tsort, type, wipe = _, date, ipairs, next, pairs, select, time, table.insert, table.sort, type, wipe

-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local Toast = LibStub("LibToast-1.0")
local L = LibStub("AceLocale-3.0"):GetLocale(addonName, false)
local BF = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceConsole-3.0", "AceEvent-3.0")

if debug then
  Broker_Followers = BF
end

local TOAST_FOLLOWER_UPGRADE = "BrokerFollowerUpgrade"
local TOAST_FOLLOWER_LEVELUP = "BrokerFollowerLevelup"


local DB_VERSION = 1
BF.followerList = {}
BF.collectedFollowersList = {}
BF.formatStrings = {
  rarityColors = {},
}
BF.threatStrings = {
  [1] = "Wild Aggression",
  [2] = "Massive Strike",
  [3] = "Group Damage",
  [4] = "Magic Debuff",
  [6] = "Danger Zones",
  [7] = "Minion Swarms",
  [8] = "Powerful Spell",
  [9] = "Deadly Minions",
  [10] = "Timed Battle",
  [11] = "Orc",
  [12] = "Ogre",
  [13] = "Undead",
  [14] = "Arakkoa",
  [15] = "Beast",
  [16] = "Demon",
  [17] = "Breaker",
  [18] = "Fury",
  [19] = "Aberration",
  [20] = "Primal",
  [21] = "Mountains",
  [22] = "Desert",
  [23] = "Snow",
  [24] = "Underground",
  [25] = "Jungle",
  [26] = "Forest",
  [27] = "Town",
  [28] = "Swamp",
  [29] = "Plains",
}

BF.specs = {
  [2] = {
    --Blood Death Knight
    1,
    2,
    7,
    8,
    10,
  },
  [3] = {
    --Frost Death Knight
    1,
    4,
    7,
    8,
    10,
  },
  [4] = {
    --Unholy Death Knight
    1,
    2,
    7,
    8,
    10,
  },
  [5] = {
    --Balance Druid
    6,
    7,
    9,
    10,
  },
  [7] = {
    --Feral Druid
    1,
    2,
    6,
    10,
  },
  [8] = {
    --Guardian Druid
    1,
    2,
    6,
    9,
  },
  [9] = {
    --Restoration Druid
    3,
    4,
    7,
    9,
  },
  [10] = {
    --Beast Master Hunter
    1,
    6,
    7,
    9,
    10,
  },
  [12] = {
    --Marksmanship Hunter
    6,
    7,
    8,
    9,
    10,
  },
  [13] = {
    --Survival Hunter
    2,
    6,
    7,
    9,
    10,
  },
  [14] = {
    --Arcane Mage
    6,
    8,
    9,
    10,
  },
  [15] = {
    --Fire Mage
    6,
    7,
    8,
    9,
    10,
  },
  [16] = {
    --Frost Mage
    2,
    7,
    8,
    9,
    10,
  },
  [17] = {
    --Brewmaster Monk
    1,
    2,
    3,
    6,
    9,
  },
  [18] = {
    --Mistweaver Monk
    3,
    4,
    6,
    8,
  },
  [19] = {
    --Windwalker Monk
    1,
    6,
    8,
    9,
    10,
  },
  [20] = {
    --Holy Paladin
    3,
    4,
    8,
    9,
  },
  [21] = {
    --Protection Paladin
    1,
    2,
    4,
    8,
    9,
  },
  [22] = {
    --Retribution Paladin
    2,
    7,
    8,
    9,
    10,
  },
  ["Discipline Priest"] = {
    --Discipline Priest
    3,
    4,
    6,
    9,
  },
  [24] = {
    --Holy Priest
    3,
    4,
    6,
    7,
  },
  [25] = {
    --Shadow Priest
    4,
    6,
    7,
    9,
    10,
  },
  [26] = {
    --Assassination Rogue
    2,
    6,
    8,
    9,
    10,
  },
  [27] = {
    --Combat Rogue
    6,
    7,
    8,
    9,
    10,
  },
  [28] = {
    --Subtlety Rogue
    2,
    6,
    7,
    8,
    9,

  },
  [29] = {
    --Elemental Shaman
    3,
    7,
    8,
    9,
    10,
  },
  [30] = {
    --Enhancement Shaman
    3,
    6,
    7,
    9,
    10,
  },
  [31] = {
    --Restoration Shaman
    3,
    4,
    7,
    8,
  },
  [32] = {
    --Affliction Warlock
    4,
    7,
    8,
    9,
    10,
  },
  [33] = {
    --Demonology Warlock
    2,
    7,
    8,
    10,
    10,
  },
  [34] = {
    --Destruction Warlock
    3,
    8,
    9,
    10,
    10,
  },
  [35] = {
    --Arms Warrior
    1,
    6,
    7,
    8,
    10,
  },
  [37] = {
    --Fury Warrior
    2,
    6,
    7,
    8,
    10,
  },
  [38] = {
    --Protection Warrior
    1,
    2,
    6,
    7,
    8,
  },
}
local defaults = {
  profile = {
    version = 1,
    autoHideDelay             = 0.5,
    availableCountersToolstip = true,
    showRerollPercent         = true,
    showNewComboPercent       = true,
    showMarkInactive          = "X",
    showInactivesList         = true,
    clickableList             = true,
    iconSize                  = 16,
    clickToShow               = 2,
    tooltipBGColor            = { r=0, g=0, b=0, a=1 },
    colorMatrix               = true,
    clickableMatrix           = true,
    onlyFillTopMatrixHalf     = false,
    showInactivesMatrix       = true,
    showToasts                = true,
    minimapIcon = {
      hide = true,
      minimapPos = 220,
      radius = 80,
    },
  },
  global = {
    version = 1,
  },
}

-- Options Handling

function BF:GenerateOptions()
  if not BF.options then

    local function newCounter()
      local i = 0
      return function( inc )
          i = i + 1
          return i
        end
    end
    local count = newCounter()

    BF.options = {
      name = addonName,
      type = 'group',
      handler = BF,
      args = {
        genconfig = {
          name = L["General Settings"],
          type = 'group',
          order = 1,
          get = function(info) return BF.db.profile[info[#info]] end,
          set = function(info, value) BF.db.profile[info[#info]] = value end,
          args = {
            clickToShow = {
              type = "select",
              style = "dropdown",
              values = {
                [1] = L["Click"],
                [2] = L["Hover"],
              },
              name = L["Show tooltip on:"],
              desc = L["Either make the tooltip pop up when you hover the icon or make it a toggle by clicking the icon"],
              order = count(),
            },
            autoHideDelay = {
              type = "range",
              min = 0,
              softMax = 3,
              name = L["Auto hide delay"],
              desc = L["When 'Show tooltip' is set to 'hover' then the tooltip stay for this amount of time after the mouse was moved away"],
              order = count(),
            },
            availableCountersToolstip = {
              type = "toggle",
              name = L["Follower Hovertooltip"],
              desc = L["When hovering over the followerporait in the follower tab or while recruiting in the inn then display small tooltip with all available counters for this follower"],
              order = count(),
            },
            showToasts = {
              type = "toggle",
              name = L["Show Toasts"],
              desc = L["Show toasts to inform you of follower level ups and upgrades. Upgrade toasts will tell you what ability/trait the follower gained."],
              order = count(),
            },
            iconSize = {
              type = "range",
              min = 4,
              max = 128,
              step = 1,
              name = L["Tooltip: Icon size"],
              order = count(),
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF.formatStrings.counterIconTolltip = "|T%s:"..value..":"..value..":0:0:64:64:5:59:5:59|t"
                BF:RedrawTooltip()
              end,
            },
            tooltipBGColor = {
              name = L["Background color"],
              desc = L["Background color"],
              type = 'color',
              order = count(),
              get  = function(info)
                return BF.db.profile[info[#info]].r, BF.db.profile[info[#info]].g, BF.db.profile[info[#info]].b, BF.db.profile[info[#info]].a
              end,
              set  = function(info, r, g, b, a)
                BF.db.profile[info[#info]] = { r = r, g = g, b = b, a = a }
                BF:RedrawTooltip()
              end,
              hasAlpha = true,
            },
            minimapIcon = {
              type = "toggle",
              name = L["Show minimap button"],
              order = count(),
              get = function(info) return not BF.db.profile.minimapIcon.hide end,
              set = function(info, value)
                BF.db.profile.minimapIcon.hide = not value
                if value then LDBIcon:Show(addonName) else LDBIcon:Hide(addonName) end
              end,
            },
            headerListViewOption = {
              type = "header",
              name = L["List view options"],
              order = count(),
            },
            clickableList=  {
              type = "toggle",
              name = L["List: Clickable cells"],
              desc = L["Clicking a followers name in the list will put it into the search box in the follower list"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
            showRerollPercent = {
              type = "toggle",
              name = L["Reroll column"],
              desc = L["In the list tooltip add a column at the front showing the % chance of getting a not-yet-acquired threat counter combination for epic followers when using a reroll certificate"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
            showNewComboPercent = {
              type = "toggle",
              name = L["Follower upgrade %"],
              desc = L["Use the Reroll column for non-epic followers to show the % of getting a new combination of threat counters"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
            showInactivesList = {
              type = "toggle",
              name = L["Show Inactive followers"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
            showMarkInactive = {
              type = "input",
              name = L["Mark inactive with"],
              desc = L["Add the following after the name of incative followers in listview"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
            headerMatrixViewOption = {
              type = "header",
              name = L["Matrix view options"],
               order = count(),
           },
            showInactivesMatrix = {
              type = "toggle",
              name = L["Count Inactives"],
              desc = L["Also count inactive followers in matrix view"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
            colorMatrix = {
              type = "toggle",
              name = L["Color followers"],
              desc = L["Color followers in the matrix view by itemlevel. (615+: Green, 630+: Blue, 645+: Epic)"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
            onlyFillTopMatrixHalf = {
              type = "toggle",
              name = L["Matrix: Top right only"],
              desc = L["Only fill the top right half of the matrix without mirroring the values onto the bottom left part as well"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
            clickableMatrix =  {
              type = "toggle",
              name = L["Matrix: Clickable cells"],
              desc = L["When a cell in the matrix is clicken it will print the list of followers that correpsond to it"],
              set = function(info, value)
                BF.db.profile[info[#info]] = value
                BF:RedrawTooltip()
              end,
              order = count(),
            },
          },
        },
      },
    }
  end
  return BF.options
end


local mechanics = {}
local q = C_Garrison.GetFollowerAbilityCounterMechanicInfo
for _, aid in pairs({11, 100, 168, 148, 160, 101, 105, 157, 122}) do
  local mid, _, tex = q(aid)
  mechanics[mid], mechanics[tex:lower():gsub("%.blp$","")] = aid, aid
end

local function getMechanicTexture(mid, formatted)
  if mechanics[mid] then
    local tex = select(3, C_Garrison.GetFollowerAbilityCounterMechanicInfo(mechanics[mid]))
    return formatted and strformat(BF.formatStrings.counterIconTolltip, tex) or tex
  end
end

local function getMechanicName(mid)
  if mechanics[mid] then
    return select(2, C_Garrison.GetFollowerAbilityCounterMechanicInfo(mechanics[mid]))
  end
end

function BF:ReleaseTooltip()
  LibQTip:Release(self.tooltip)
  self.tooltip.Columns = nil
  self.tooltip.Rows = nil
  self.tooltip = nil
end

function BF:PrepareTooltip(anchorFrame, tableInsteadOfList)
  local tooltip = LibQTip:Acquire("Broker_Followers_Tooltip", 1, "LEFT")
  if tooltip:IsShown() then
    tableInsteadOfList = self.matrixIsUsed
  end
  self.tooltip = tooltip
  self:DrawTooltip(tooltip, tableInsteadOfList)
  tooltip:SmartAnchorTo(anchorFrame)
  tooltip:Show()
end

function BF:RedrawTooltip()
  -- Check if we actually view matrix tooltip
  local tooltip = LibQTip:Acquire("Broker_Followers_Tooltip", 1, "LEFT")
  -- Only redraw if the tooltip is shown
  if not tooltip:IsShown() then return end

  self.tooltip = tooltip
  self:DrawTooltip(tooltip, self.matrixIsUsed)
end

function BF:DrawTooltip(tooltip, tableInsteadOfList)
  self:UpdateFollowerCounters()
  self:UpdateRerollCombos()

  tooltip:Clear()

  tooltip:SetBackdropColor(self.db.profile.tooltipBGColor.r, self.db.profile.tooltipBGColor.g, self.db.profile.tooltipBGColor.b, self.db.profile.tooltipBGColor.a)
  tooltip.Rows    = {}
  local function sortFollowers(f1, f2)
    local f1a, f2a = BF.collectedFollowersList[f1.followerID], BF.collectedFollowersList[f2.followerID]
    if f1.isCollected and f2.isCollected  then
      for _,i in pairs({1,2,3,4,6,7,8,9,10}) do
      --for i=1,10 do
        if f1a[i] and not f2a[i] then
          return true
        end
        if not f1a[i] and f2a[i] then
          return false
        end
      end
      return f1.name < f2.name
    else
      if not f1.isCollected and not f2.isCollected then
        return f1.name > f2.name
      else
        return f1.isCollected
      end
    end
  end

  local function sortAbilities(a1, a2)
    if not a1.isTrait and a2.isTrait  then
      return true
    end
    if a1.isTrait and  not a2.isTrait  then
      return false
    end
    local a1c = next(a1.counters)
    local a2c = next(a2.counters)
    if a1.isTrait and a2.isTrait  then
      if a1c or a2c then
        return (a1c or 0) > (a2c or 0)
      else
        return a1.name > a2.name
      end
    end
    a1c = next(a1.counters)
    a2c = next(a2.counters)
    return a1c < a2c
  end
  if not tooltip.Columns then
    tooltip.Columns = {}
  end
  if not tooltip.Rows then
    tooltip.Rows = {}
  end
  local function MyAddColumn(nameType, displayName)
    if nameType then
      if not displayName then
        BF:Debug("MyAddColumn: missing displayName for", nameType)
        displayName = nameType
      end
      if not formatFunc then
        --formatFunc = CellFormat.Default
      end

      if not tooltip.Columns[nameType] then
        local colindex = tooltip:AddColumn("CENTER")
        tooltip:SetCell(1, colindex, displayName)

        tooltip.Columns[nameType] = { col = colindex }
      else
        BF:Debug("a column already exists for", nameType)
      end
    end
  end
  local function AddSpanner(name, abbr, first, last)
    -- TODO - can we, from the size of span and maybe the character count of
    -- name make a guess as to whether we could use the name instead of Abbr?
    if first <= last then
      tooltip:SetCell(1, first, abbr, nil, nil, 1 + last - first)
    end
  end
  local function openFollowerTabSearch( tooltip_cell, searchTerm )
    local searchBox = GarrisonMissionFrameFollowers.SearchBox:IsVisible() and GarrisonMissionFrameFollowers.SearchBox or GarrisonLandingPage.FollowerList.SearchBox:IsVisible() and GarrisonLandingPage.FollowerList.SearchBox
    if searchBox then
      searchBox:SetText(searchTerm)
      searchBox.clearText = searchTerm
    end
  end
  local function displayFollowersWithCounter(tooltip_cell, counters )
    local first, second = math.min(counters[1],counters[2]), math.max(counters[1],counters[2])
    local list = tooltip.counterDualSum[first][second].list
    if #list > 0 then
      local formattedList = {}
      for _, followerID in pairs(list) do
        tinsert(formattedList, strformat("%s (%s)", C_Garrison.GetFollowerLink(followerID), BF:GetRerollPercent( followerID )))
        --strformat(self.formatStrings.counterIconTolltip, GetItemIcon(118354))
      end
      local displayString =  getMechanicTexture(first, true).."/"..
                             getMechanicTexture(second, true)..": "..      
                             table.concat(formattedList,", ")
      BF:Print(displayString)
    end
    --strformat(self.formatStrings.counterIconTolltip, GetItemIcon(118354))
    --C_Garrison.GetFollowerName
  end 

  colors = {
    green = {r=0, g=1, b=0, a=1},
    white = {r=1, g=1, b=1, a=1},
    red = {r=1, g=0, b=0, a=1},

    lightGray = {r=(108/255), g=(115/255), b=(120/255), a=1},
    green = {r=(32/255), g=(255/255), b=(32/255), a=1},
    red = {r=(255/255), g=(25/255), b=(25/255), a=1},
    yellow = {r=(255/255), g=(228/255), b=(0/255), a=1},

    --lightGray = {r=0.25, g=0.25, b=0.25, a=1},
    darkGray = {r=0.1, g=0.1, b=0.1, a=1},
    lineGrey = {r=(108/255), g=(115/255), b=(120/255), a=1},
    --lightGray = {r=(82/255), g=(91/255), b=(97/255), a=1},
  }
  local function AddSeparator(tooltip)
    tooltip:AddSeparator(1, colors.lineGrey.r, colors.lineGrey.g, colors.lineGrey.b, colors.lineGrey.a)
  end

  local globaldb = self.db.global
  local profiledb = self.db.profile

  if tableInsteadOfList then
    self.matrixIsUsed = true
    -- 9x9 table, each cell showing the number of follower who counters both
    -- threats from column and row

    if tooltip:GetLineCount() < 3 then
      tooltip.Rows["icon"] = tooltip.Rows["icon"] or tooltip:AddHeader()
      tooltip.Rows["sum"]  = tooltip.Rows["sum"]  or tooltip:AddHeader()
    end
    MyAddColumn("icon","")

    -- Setup the tooltip: Add the textures to the cells of the irst row and
    -- column, then "" all the other cells so we start with a clean matrix
    for k=1,10 do
      local tex = getMechanicTexture(k, true)
      if tex then
        local row = tooltip.Rows[k] or tooltip:AddHeader()
        --AddSeparator(tooltip)
        MyAddColumn(tostring(k))
        tooltip:SetCell( 1, tooltip.Columns[tostring(k)].col, tex)
        tooltip:SetCell( row, 1, tex)
        tooltip.Rows[k] = row
        for j=1,10 do
          if tooltip.Columns[tostring(j)]then
            tooltip:SetCell( tooltip.Rows[k], tooltip.Columns[tostring(j)].col, "")
          end
        end
      end
    end
    local counterSum = {}
    tooltip.counterDualSum = {}
    local function addDualCounter( c1, c2, iLevel, followerData)
      local first, second = math.min(c1,c2), math.max(c1,c2)
      local quality = 1
      if iLevel >= 615 then
        quality = 2
      end
      if iLevel >= 630 then
        quality = 3
      end
      if iLevel >= 645 then
        quality = 4
      end
      BF:Debug( c1, c2, iLevel, quality, followerData.name)
      tooltip.counterDualSum[first]     = tooltip.counterDualSum[first] or {}
      tooltip.counterDualSum[first][second] = tooltip.counterDualSum[first][second] or {}
      tooltip.counterDualSum[first][second].num     = ( tooltip.counterDualSum[first][second].num or 0 ) + 1
      tooltip.counterDualSum[first][second].quality = math.max( tooltip.counterDualSum[first][second].quality or 0,  quality )
      tooltip.counterDualSum[first][second].list    = tooltip.counterDualSum[first][second].list or {}
      tinsert(tooltip.counterDualSum[first][second].list, followerData.followerID)
    end


    table.sort(self.followerList, sortFollowers)
    for _, followerData in pairs( self.followerList ) do
      if followerData.isCollected and ( followerData.status ~= "Inactive" or self.db.profile.showInactivesMatrix ) then
        local abilities = C_Garrison.GetFollowerAbilities(followerData.followerID)
        local counters = {}
        for id, abilityData in pairs( abilities ) do

          local counterID, _ = next(abilityData.counters)
          if not abilityData.isTrait then
            table.insert( counters, counterID)
            counterSum[counterID] = ( counterSum[counterID] or 0 ) + 1
            tooltip:SetCell(2, tooltip.Columns[tostring(counterID)].col, counterSum[counterID])
            if self.db.profile.clickableMatrix then
              tooltip:SetCellScript(2, tooltip.Columns[tostring(counterID)].col, "OnMouseUp", openFollowerTabSearch, getMechanicName(counterID))
            end

          end
        end
        if #counters > 1 then
          addDualCounter( counters[1], counters[2], followerData.iLevel, followerData )
        end
      end
    end

    local r,c
    for r, cData in pairs(tooltip.counterDualSum) do
      for c, rData in pairs(cData) do
        local formatedNumber
        if self.db.profile.colorMatrix then
          formatedNumber = strformat(self.formatStrings.rarityColors[rData.quality], rData.num)
        else
          formatedNumber = rData.num
        end
        tooltip:SetCell( tooltip.Rows[r], tooltip.Columns[tostring(c)].col, formatedNumber)
        if self.db.profile.clickableMatrix then
          tooltip:SetCellScript(tooltip.Rows[r], tooltip.Columns[tostring(c)].col, "OnMouseUp", displayFollowersWithCounter, {c, r})
        end
        if not self.db.profile.onlyFillTopMatrixHalf then
          -- Mirror the matrix
          tooltip:SetCell( tooltip.Columns[tostring(c)].col, tooltip.Rows[r], formatedNumber)
          if self.db.profile.clickableMatrix then
            tooltip:SetCellScript(tooltip.Columns[tostring(c)].col, tooltip.Rows[r], "OnMouseUp", displayFollowersWithCounter, {r, c})
          end
        end
      end
    end
  else
    self.matrixIsUsed = false
    -- Tooltip: List of followers with 5 columns for counters/traits
    tooltip:AddHeader()
    table.sort( self.followerList, sortFollowers)

    if self.db.profile.showRerollPercent then
      local rerollIcon =  strformat(self.formatStrings.counterIconTolltip, GetItemIcon(118354))
      MyAddColumn("reroll")
      tooltip:SetCell(1, tooltip.Columns["reroll"].col, rerollIcon)
    end

    MyAddColumn("counter1","")
    MyAddColumn("counter2","")
    AddSpanner("Threats", "vs", tooltip.Columns["counter1"].col, tooltip.Columns["counter2"].col)
    MyAddColumn("trait1","")
    MyAddColumn("trait2","")
    MyAddColumn("trait3","")
    AddSpanner("Traits", "Traits", tooltip.Columns["trait1"].col, tooltip.Columns["trait3"].col)

    for _, followerData in pairs(  self.followerList ) do
      if followerData.isCollected and
        (self.db.profile.showInactivesList or followerData.status ~= "Inactive") then
        local row = tooltip:AddLine("")
        local coloredName = strformat(self.formatStrings.rarityColors[followerData.quality], followerData.name)
        tooltip:SetCell(row, 1, strformat("%s%s (%d)", (followerData.status == "Inactive" and self.db.profile.showMarkInactive or ""), coloredName, followerData.level == 100 and followerData.iLevel or followerData.level) )
        if self.db.profile.clickableList then
          tooltip:SetCellScript(row, 1, "OnMouseUp", openFollowerTabSearch, followerData.name)
        end
        if self.db.profile.showRerollPercent and
           ( ( followerData.combo and self.reRollCombos[followerData.combo] > 1 ) or
               self.db.profile.showNewComboPercent )then
          local reroll = self:GetRerollPercent( followerData.followerID )
          tooltip:SetCell(row, tooltip.Columns["reroll"].col, reroll)
        end
        local counter, trait = 1,1
        local abilities = C_Garrison.GetFollowerAbilities(followerData.followerID)
        table.sort(abilities, sortAbilities)

        for id, abilityData in pairs( abilities ) do
          local _, counterData = next(abilityData.counters)

          if  abilityData.isTrait then
            tooltip:SetCell(row, tooltip.Columns["trait"..trait].col, strformat(self.formatStrings.counterIconTolltip, counterData and counterData.icon or abilityData.icon))
            trait = trait+1
          else
            tooltip:SetCell(row, tooltip.Columns["counter"..counter].col, strformat(self.formatStrings.counterIconTolltip, counterData and counterData.icon or abilityData.icon) )
            counter = counter+1
          end
        end
      end
    end
  end
end

function BF:GetRerollPercent( passedID, isSpecID  )
  -- It's either a followerID or directly the specID
  local specID = isSpecID and passedID or C_Garrison.GetFollowerClassSpec(passedID)
  local spec = self.specs[specID] or self.specs[specID] or nil
  local allCombos, wantedCombos = 0, 0
  if spec then
    for i=1,(#spec-1) do
      for j=i+1,#spec do
        local combo = spec[i] + spec[j]/100
        allCombos = allCombos +1
        if not self.reRollCombos[combo] then
          wantedCombos = wantedCombos+1
        end
      end
    end
  else
    BF:Debug("ERROR", "GetRerollPercent", C_Garrison.GetFollowerName(followerID), specID)
  end
  local percentage = ( allCombos==0 ) and 0 or (  wantedCombos / allCombos * 100 )
  return math.floor(percentage).."%"
end

--

function BF:FollowerXPChange(event, followerID, xpAward, oldXP, oldLevel, oldQuality)
  self.Debug(event, followerID, xpAward, oldXP, oldLevel, oldQuality)
  local name, displayID, level, quality, currXP, maxXP = C_Garrison.GetFollowerMissionCompleteInfo(followerID)
  self.Debug(name, displayID, xpAward, oldXP, level, quality)

  if xpAward > 0 then
    -- XP gained
  end
  if oldLevel ~= level then
    -- Levelup
    local replacements = {
      ["lvl"]  = level,
      ["name"] = strformat(self.formatStrings.rarityColors[quality],name),
    }
    local levelupString = string.gsub(L["%name% is now level %lvl%"], "%%(%w+)|?([^%%]*)%%",  function (s, default)
      local key = tostring(s)
      return replacements[key] or default
    end)

    Toast:Spawn( TOAST_FOLLOWER_LEVELUP, levelupString )

    self:RedrawTooltip()
  end
  if oldQuality ~= quality then
    -- Qualitiy upgrade
    local formatedToastStrings  = {}
    local followerAbilitiesData = C_Garrison.GetFollowerAbilities(followerID)

    local function getAcquiredString( abilityIndex )
      local replacements, icon= {
        ["name"] = followerAbilitiesData[abilityIndex].name,
      }
      if followerAbilitiesData[abilityIndex].isTrait then
        if not followerAbilitiesData[abilityIndex].icon then
          self:UpdateFollowerCounters()
          self:UpdateRerollCombos()
        end
        replacements["icon"] = strformat(self.formatStrings.counterIconTolltip, followerAbilitiesData[abilityIndex].icon)
      else
        if not counterID then
          self:UpdateFollowerCounters()
          self:UpdateRerollCombos()
        end
        local counterID = next( followerAbilitiesData[abilityIndex].counters)
        replacements["icon"] = getMechanicTexture(counterID, true)
      end

      local text = string.gsub(L["Acquired %name% (%icon%)"], "%%(%w+)|?([^%%]*)%%",  function (s, default)
        local key = tostring(s)
        return replacements[key] or default
      end)
      return text
    end


    local formatedToastStrings = {}
    local counters = {}
    tinsert(counters, (next(followerAbilitiesData[1].counters)))
    -- Get icon for the ability the follower had since beginning
    if not counter[1] then
      self:UpdateFollowerCounters()
      self:UpdateRerollCombos()
    end
    local replacements = {
      ["name"] = strformat(self.formatStrings.rarityColors[quality], name),
      ["icon"] = getMechanicTexture(counters[1], true)
    }
    local firstTextLine = string.gsub("%name% (%icon%)", "%%(%w+)|?([^%%]*)%%",  function (s, default)
      local key = tostring(s)
      return replacements[key] or default
    end)
    -- Name
    tinsert(formatedToastStrings, firstTextLine)
    if quality == 3 then
      -- Green -> Blue
      -- Gained C_Garrison.GetFollowerAbilities(followerID)[3]
      
      tinsert(formatedToastStrings, getAcquiredString(3))
    elseif quality == 4 then
      -- Blue -> Epic
      -- Gained C_Garrison.GetFollowerAbilities(followerID)[4] + [5]
      -- Determine if we got a new combination of counters
      -- self.reRollCombos
      tinsert(counters, (next(followerAbilitiesData[4].counters)))
      table.sort(counters)
      local combo = counters[1]+counters[2]/100
      tinsert(formatedToastStrings, getAcquiredString(4))
      if not self.reRollCombos[combo] then
        tinsert(formatedToastStrings, strformat( L["New epic combo: %s/%s"],
          getMechanicTexture(counters[1], true) or "ERROR",
          getMechanicTexture(counters[2], true) or "ERROR"
        ))
      end
      tinsert(formatedToastStrings, getAcquiredString(5))
    end
    Toast:Spawn( TOAST_FOLLOWER_UPGRADE, table.concat(formatedToastStrings, "\n"))
    self:RedrawTooltip()
  end
end

local function toastFollowerUpgrade(toast, text)
  toast:SetTitle(L["Follower upgrade"])
  toast:SetText(text)
end

local function toastFollowerLevelup(toast, text)
  toast:SetTitle(L["Follower levelup"])
  toast:SetText(text)
end

function BF:OnEnable()
  self:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED",  "FollowerXPChange")

  self:RegisterEvent("GARRISON_FOLLOWER_ADDED",       "RedrawTooltip")
  self:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE", "RedrawTooltip")
  self:RegisterEvent("GARRISON_FOLLOWER_REMOVED",     "RedrawTooltip")
  self:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE", "RedrawTooltip")
  self:RegisterEvent("GARRISON_UPDATE",               "UpdateTheFuckingFollowers")
  self:RegisterEvent("GARRISON_MISSION_NPC_OPENED",   "UpdateTheFuckingFollowers")
end

function BF:OnDisable()
  self:UnregisterEvent("GARRISON_FOLLOWER_XP_CHANGED"   )
  self:UnregisterEvent("GARRISON_FOLLOWER_ADDED"        )
  self:UnregisterEvent("GARRISON_FOLLOWER_LIST_UPDATE"  )
  self:UnregisterEvent("GARRISON_FOLLOWER_REMOVED"      )
  self:UnregisterEvent("GARRISON_UPDATE"                )
  self:UnregisterEvent("GARRISON_MISSION_NPC_OPENED"    )
end

---- Parse Follower List

function BF:UpdateTheFuckingFollowers()
  self:Debug("GARRISON_UPDATE")
  self:UpdateFollowerCounters()
  self:UpdateRerollCombos()
end

function BF:UpdateFollowerCounters()
  local followerData, counterID, followerAbilitiesData
  self.followerList = C_Garrison.GetFollowers()
  self.collectedFollowersList = {}
  for _, followerData in pairs( self.followerList ) do
    self.collectedFollowersList[followerData.followerID] = {}

    for _, followerAbilitiesData in ipairs(C_Garrison.GetFollowerAbilities(followerData.followerID)) do
      counterID = next( followerAbilitiesData.counters)
      if counterID then
        self.collectedFollowersList[followerData.followerID][counterID] = true
      end
    end
  end
end

function BF:UpdateRerollCombos()
  local collected = 0
  for _, followerData in pairs( self.followerList ) do
    if followerData.isCollected or true then
      collected = collected +1
      local abilities = C_Garrison.GetFollowerAbilities(followerData.followerID)
      local counters = {}
      for id, abilityData in pairs( abilities ) do
        local counterID, _ = next(abilityData.counters)
        if not abilityData.isTrait then
          table.insert( counters, counterID)
        end
      end
      if #counters == 2 then
        table.sort(counters)
        local combo = counters[1]+counters[2]/100
        followerData.combo  = combo
        self.reRollCombos[combo] = ( self.reRollCombos[combo] or 0 ) +1
      end
    end
  end
end

---- Enhancing the Inn Recruitment

function BF:EnhanceRecruitFrame()
  --if true then return end
  self:UpdateFollowerCounters()
  self:UpdateRerollCombos()
  local recruitFrame = GarrisonRecruitSelectFrame.FollowerSelection
  local followers = C_Garrison.GetAvailableRecruits();
  for i=1, 3 do
    recruitFrame["Recruit"..i].PortraitFrame:SetScript("OnEnter",
      function(tt)
        if not BF.db.profile.availableCountersToolstip then return end
        local tooltip = LibQTip:Acquire("Broker_Followers_EnhancementTooltip", 1, "LEFT")
        BF.enhanceTooltip = tooltip
        BF:AdditionalAbilitesTooltip(tooltip, followers[i])
        --tooltip:SmartAnchorTo(tt)
        tooltip:SetPoint("RIGHT",tt,"LEFT")
        tooltip:Show()
      end
    )
    recruitFrame["Recruit"..i].PortraitFrame:SetScript("OnLeave",
      function()
        if not BF.db.profile.availableCountersToolstip then return end
        LibQTip:Release(BF.enhanceTooltip)
        BF.enhanceTooltip.Columns = nil
        BF.enhanceTooltip.Rows = nil
        BF.enhanceTooltip = nil
      end
    )
  end
end

hooksecurefunc("GarrisonRecruitSelectFrame_UpdateRecruits", function(waiting)
  if not waiting then
    BF:EnhanceRecruitFrame()
  end
end)


function BF:EnhanceFollowerTab( parentFrame )
  local followerTabPortrait = parentFrame.FollowerTab.PortraitFrame
  followerTabPortrait:SetScript("OnEnter",
    function(tt)
      local tooltip = LibQTip:Acquire("Broker_Followers_EnhancementTooltip", 1, "LEFT")
      BF.enhanceTooltip = tooltip
      BF:AdditionalAbilitesTooltip(tooltip, parentFrame.FollowerTab.classSpec)
      tooltip:SetPoint("RIGHT",tt,"LEFT")
      tooltip:Show()
    end)
  followerTabPortrait:SetScript("OnLeave",
    function()
    LibQTip:Release(BF.enhanceTooltip)
    BF.enhanceTooltip = nil
  end )
end

function BF:AdditionalAbilitesTooltip( tooltip, specID)
  local possibleAbilities = self.specs[specID]
  for _, aid in pairs(possibleAbilities) do
    local tex = getMechanicTexture( aid, true )
    tooltip:AddHeader(tex)
  end
  tooltip:AddHeader( self:GetRerollPercent( specID, true ) )
end

---- Database Maintenance

function BF:MigrateSavedData()
  local  globaldb = self.db.global
  local profiledb = self.db.profile

  -- all db sections are current we can exit immediately
  if profiledb.version == DB_VERSION and
      globaldb.version == DB_VERSION then
    return
  else
    BF:Print(L["DB_VERSION_UPGRADE_PURGE"])
  end

  -- Migrating from before addon began keeping track of database changes
  --[[
  if globaldb.version < X then
    -- Upgrade from before version X
  end
  if profiledb.version < X then
    -- Upgrade from before version X
  end
  --]]--

  profiledb.version = DB_VERSION
  globaldb.version = DB_VERSION
end

function BF:OnInitialize() -- ADDON_LOADED, only saved variables are guaranteed to be present
  UIParentLoadAddOn("Blizzard_GarrisonUI")
  self.db = LibStub("AceDB-3.0"):New("Broker_Follower_DB", defaults, true)
  self:MigrateSavedData()

  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, self.GenerateOptions)
  self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName)

  self.BFLDB = ldb:NewDataObject(addonName, {
    type = "data source",
    text = "",
    icon = "Interface\\Icons\\inv_misc_lockboxghostiron",
    OnClick = function(clickedframe, button)
      if button == "RightButton" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
      else
        -- 1 = click2show
        -- 2 = hover2show
        if self.db.profile.clickToShow == 2 then
          self.db.profile.showInactivesMatrix = not self.db.profile.showInactivesMatrix
          self:PrepareTooltip(clickedframe, not IsShiftKeyDown())
        else
          -- Click 2 Show handling
          local tooltip = LibQTip:Acquire("Broker_Followers_Tooltip", 1, "LEFT")
          self.tooltip = tooltip
          self.tooltip:SetAutoHideDelay( nil )          
          if tooltip:IsShown() then
            -- Tooltip is shown, hide it on click
            self:ReleaseTooltip()
          else
            -- No tooltip, show it on click
            self:PrepareTooltip(clickedframe, not IsShiftKeyDown())
          end
        end
      end
    end,
    OnEnter = function(enteredFrame)
      -- 1 = click2show
      -- 2 = hover2show
      if self.db.profile.clickToShow == 1 then return end
      self:PrepareTooltip(enteredFrame, not IsShiftKeyDown())
      self.tooltip:SetAutoHideDelay( self.db.profile.autoHideDelay or 0.25, enteredFrame )
      self.tooltip.OnRelease = function() BF:ReleaseTooltip() end
    end,
    OnLeave = function(tt)
      -- 1 = click2show
      -- 2 = hover2show
      if self.db.profile.clickToShow == 1 or true then return end
      LibQTip:Release(self.tooltip)
      if self.tooltip then
        self.tooltip.Columns = nil
        self.tooltip.Rows = nil
        self.tooltip = nil
      end
    end,
  })

  LDBIcon:Register(addonName, self.BFLDB, self.db.profile.minimapIcon)

  self.reRollCombos = {}
  self:EnhanceFollowerTab( GarrisonMissionFrame )
  self:EnhanceFollowerTab( GarrisonLandingPage )
  Toast:Register(TOAST_FOLLOWER_UPGRADE, toastFollowerUpgrade)
  Toast:Register(TOAST_FOLLOWER_LEVELUP, toastFollowerLevelup)

  -- Update the .format strings
  self.formatStrings.counterIconTolltip = "|T%s:"..self.db.profile.iconSize..":"..self.db.profile.iconSize..":0:0:64:64:5:59:5:59|t"
  for i=0,7 do
    self.formatStrings.rarityColors[i] = "|c"..select(4,GetItemQualityColor(i)).."%s|r"
  end
end



function BF:Debug(...)
  DebugPrint( ... )
end